<?php
// This file is part of the Studyplan plugin for Moodle
//
// Moodle is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Moodle is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Moodle.  If not, see <https://www.gnu.org/licenses/>.
/**
 * Report 
 * @package    local_treestudyplan
 * @copyright  2023 P.M. Kuipers
 * @license    https://www.gnu.org/copyleft/gpl.html GNU GPL v3 or later
 */

require_once("../../config.php");

use local_treestudyplan\contextinfo;
use \local_treestudyplan\courseservice;
use \local_treestudyplan\studyplan;
use \local_treestudyplan\studyplanpage;
use \local_treestudyplan\premium;

require_once($CFG->libdir.'/weblib.php');

$PAGE->set_url("/local/treestudyplan/studyplan-report.php", array());
require_login();

// Figure out the context (category or system, based on either category or context parameter).
$pageid = required_param('page', PARAM_INT); // Category id.
$page = studyplanpage::find_by_id($pageid);
$studyplan = $page->studyplan();
$context = $studyplan->context();
$ci = new contextinfo($context);
$contextname = $ci->pathstr();

$firstperiod = optional_param('firstperiod', 0, PARAM_INT); // First period to show
$lastperiod = optional_param('lastperiod', 0, PARAM_INT); // Last periode to show



$PAGE->set_pagelayout('report');
$PAGE->set_title(get_string('studyplan_report', 'local_treestudyplan'));
$PAGE->set_heading(get_string('studyplan_report', 'local_treestudyplan'));

if (!premium::enabled()) {
    throw new \moodle_exception("error:nopremiumaccess","local_treestudyplan","",premium::statusdescription());
}

if(!has_capability('local/treestudyplan:viewuserreports', $context)) {
    throw new \moodle_exception("error:nostudyplanviewaccess","local_treestudyplan","",$contextname);
}

// Load javascripts and specific css.
$PAGE->requires->css(new moodle_url($CFG->wwwroot.'/local/treestudyplan/css/bootstrap-vue/bootstrap-vue.css'));
if ($CFG->debugdeveloper) {
    $PAGE->requires->css(new moodle_url($CFG->wwwroot.'/local/treestudyplan/css/devstyles.css'));
}
$PAGE->requires->js_call_amd('local_treestudyplan/page-result-overview', 'init', [$studyplan->id(), $page->id(), $firstperiod, $lastperiod]);


/**
 * Shortcut function to provide translations
 *
 * @param mixed $str Translation key
 * @param null|string[] $param Parameters to pass to translation
 * @param string $plugin Location to search for translation strings
 * @return string Translation of key
 */
function t($str, $param = null, $plugin = 'local_treestudyplan') {
    print get_string($str, $plugin, $param);
}

print $OUTPUT->header();
print "<h3><b>{$contextname}</b> / {$studyplan->name()}</h3>"; 
if($studyplan->name() != $page->fullname()) {
    print "<h4>{$page->fullname()}</h4>";
}
?>
<div id='root'>
    <div class='vue-loader' v-show='false'>
        <div class="spinner-border text-primary" role="status">
            <span class="sr-only">Loading...</span>
        </div>
    </div>
    <div v-cloak>
        <div class='vue-loader' v-if="!structure">
            <div class="spinner-border text-primary" role="status">
                <span class="sr-only">Loading...</span>
            </div>
        </div>
        <template v-else>
            <div class='container q-pageperiodselection'>
                <div class='row'>
                    <div class="col-sm-2 q-label">{{text.page}}</div>
                    <div class="col-sm-6">
                        <select class="custom-select" :value="page.id" @change="selectedPage">
                            <template v-if="studyplan">
                                <option v-for="p in studyplan.pages" 
                                    :key="p.id" 
                                    :value="p.id"
                                    :selected="(page.id == p.id)?true:false"
                                    >{{p.fullname}}</option>
                            </template>
                            <option v-else disabled>{{text.loading}}</option>
                        </select>
                    </div>
                </div>
                <div class='row'>
                    <div class="col-sm-2 q-label" >{{text.period}} ({{text.from}} / {{text.to}})</div>
                    <div class="col-sm-2">
                        <select class="custom-select" @change="selectedFirstPeriod">
                            <template v-if="page">
                                <option v-for="p in page.perioddesc" 
                                    :key="p.id" 
                                    :value="p.period"
                                    :selected="(structure.firstperiod == p.period)?true:false"
                                    >{{p.fullname}}</option>
                            </template>
                            <option v-else disabled>{{text.loading}}</option>
                        </select>
                    </div>
                    <div class="col-sm-2 ">
                        <select class="custom-select" @change="selectedLastPeriod">
                            <template v-if="page">
                                <template v-for="p in page.perioddesc">
                                    <option v-if="p.period >= structure.firstperiod"
                                        :key="p.id" 
                                        :value="p.period"
                                        :selected="(structure.lastperiod == p.period)?true:false"
                                        >{{p.fullname}}</option>
                                </template>
                            </template>
                            <option v-else disabled>{{text.loading}}</option>
                        </select>
                    </div>
                </div>
            </div>
            <q-studyplanreport :structure="structure" ></s-studyplanreport>
        </template>
    </div>
</div>
<?php

print $OUTPUT->footer();
